@@ -31,8 +31,7 @@
 # include <config.h>
 #endif
 
-#include "../libelf/libelfP.h"
-#undef	_
+#include "libelfP.h"
 #include "libdwflP.h"
 
 #if !USE_BZLIB
@@ -152,7 +151,7 @@ libdw_open_elf (int *fdp, Elf **elfp, bool close_on_fail, bool archive_ok,
 	  elf->state.ar.elf_ar_hdr.ar_name = "libdwfl is faking you out";
 	  elf->state.ar.elf_ar_hdr.ar_size = elf->maximum_size - offset;
 	  elf->state.ar.offset = offset - sizeof (struct ar_hdr);
-	  Elf *subelf = elf_begin (-1, ELF_C_READ_MMAP_PRIVATE, elf);
+	  Elf *subelf = elf_begin (-1, elf->cmd, elf);
 	  elf->kind = ELF_K_NONE;
 	  if (unlikely (subelf == NULL))
 	    error = DWFL_E_LIBELF;
@@ -206,14 +205,16 @@ __libdw_open_elf_memory (char *data, size_t size, Elf **elfp, bool archive_ok)
 {
   /* It is ok to use `fd == -1` here, because libelf uses it as a value for
      "no file opened" and code supports working with this value, and also
-     `never_close_fd == false` is passed to prevent closing non-existant file.
+     `never_close_fd == false` is passed to prevent closing non-existent file.
      The only caveat is in `decompress` method, which doesn't support
      decompressing from memory, so reading compressed zImage using this method
      won't work.  */
   int fd = -1;
   *elfp = elf_memory (data, size);
-  /* Allow using this ELF as reference for subsequent elf_begin calls.  */
-  (*elfp)->cmd = ELF_C_READ_MMAP_PRIVATE;
+  if (unlikely(*elfp == NULL))
+    {
+      return DWFL_E_LIBELF;
+    }
   return libdw_open_elf (&fd, elfp, false, archive_ok, true, false, true);
 }
 
