@@ -32,7 +32,6 @@
 #include <gelf.h>
 #include <libasm.h>
 #include <libebl.h>
-#include <libintl.h>
 
 
 /* Backend handle.  */
@@ -86,11 +85,6 @@ struct ebl
 typedef Ebl *(*ebl_bhinit_t) (Elf *, GElf_Half, Ebl *);
 
 
-/* gettext helper macros.  */
-#undef _
-#define _(Str) dgettext ("elfutils", Str)
-
-
 /* LEB128 constant helper macros.  */
 #define ULEB128_7(x)	(BUILD_BUG_ON_ZERO ((x) >= (1U << 7)) + (x))
 
