@@ -23,7 +23,6 @@
 #include <argp.h>
 #include <fcntl.h>
 #include <inttypes.h>
-#include <libintl.h>
 #include <locale.h>
 #include <stdbool.h>
 #include <stdio.h>
@@ -581,12 +580,12 @@ show_full_content (Ebl *ebl, const char *fname, uint32_t shstrndx)
 		printf ("%02hhx%02hhx%02hhx%02hhx ",
 			cp[inner], cp[inner + 1], cp[inner + 2],
 			cp[inner + 3]);
-	      fputc_unlocked (' ', stdout);
+	      fputc (' ', stdout);
 
 	      for (size_t inner = 0; inner < 16; ++inner)
-		fputc_unlocked (isascii (cp[inner]) && isprint (cp[inner])
+		fputc (isascii (cp[inner]) && isprint (cp[inner])
 				? cp[inner] : '.', stdout);
-	      fputc_unlocked ('\n', stdout);
+	      fputc ('\n', stdout);
 	    }
 
 	  printf (" %04zx ", cnt);
@@ -602,14 +601,14 @@ show_full_content (Ebl *ebl, const char *fname, uint32_t shstrndx)
 
 	  for (inner = 2 * (16 - inner) + (16 - inner + 3) / 4 + 1; inner > 0;
 	       --inner)
-	    fputc_unlocked (' ', stdout);
+	    fputc (' ', stdout);
 
 	  for (inner = 0; inner < remaining; ++inner)
-	    fputc_unlocked (isascii (cp[inner]) && isprint (cp[inner])
+	    fputc (isascii (cp[inner]) && isprint (cp[inner])
 			    ? cp[inner] : '.', stdout);
-	  fputc_unlocked ('\n', stdout);
+	  fputc ('\n', stdout);
 
-	  fputc_unlocked ('\n', stdout);
+	  fputc ('\n', stdout);
 	}
     }
 
@@ -641,12 +640,12 @@ disasm_output (char *buf, size_t buflen, void *arg)
     printf ("%8" PRIx64 ":   ", (uint64_t) info->addr);
 
   if (info->bytes_color != NULL)
-    fputs_unlocked (info->bytes_color, stdout);
+    fputs (info->bytes_color, stdout);
   size_t cnt;
   for (cnt = 0; cnt < (size_t) MIN (info->cur - info->last_end, 8); ++cnt)
     printf (" %02" PRIx8, info->last_end[cnt]);
   if (info->bytes_color != NULL)
-    fputs_unlocked (color_off, stdout);
+    fputs (color_off, stdout);
 
   printf ("%*s %.*s\n",
 	  (int) (8 - cnt) * 3 + 1, "", (int) buflen, buf);
@@ -664,12 +663,12 @@ disasm_output (char *buf, size_t buflen, void *arg)
 	printf ("%8" PRIx64 ":   ", (uint64_t) info->addr);
 
       if (info->bytes_color != NULL)
-	fputs_unlocked (info->bytes_color, stdout);
+	fputs (info->bytes_color, stdout);
       for (; cnt < (size_t) (info->cur - info->last_end); ++cnt)
 	printf (" %02" PRIx8, info->last_end[cnt]);
       if (info->bytes_color != NULL)
-	fputs_unlocked (color_off, stdout);
-      putchar_unlocked ('\n');
+	fputs (color_off, stdout);
+      putchar ('\n');
       info->addr += info->cur - info->last_end - 8;
     }
 
